import 'dart:convert';

import 'package:http/http.dart' as http;
import 'package:json_api_document/json_api_document.dart';

class JsonApiClient {
  JsonApiClient({
    this.baseUrl = '',
    ClientFactory clientFactory,
    Map<String, String> headers = const {},
    String charset = 'utf-8',
  })  : clientFactory = clientFactory ?? (() => http.Client()),
        defaultHeaders = Map.unmodifiable({}
          ..addAll(headers)
          ..['Accept'] = Document.mediaType);

  final String baseUrl;
  final ClientFactory clientFactory;
  final Map<String, String> defaultHeaders;

  Future<Document> fetch(String url,
          {Map<String, String> headers = const {}}) =>
      _exec((_) => _.get('${baseUrl}${url}', headers: _makeHeaders(headers)));

  Future<Document> create(String url, Document doc,
          {Map<String, String> headers = const {}}) =>
      _exec((_) => _.post('${baseUrl}${url}',
          body: json.encode(doc), headers: _makeHeaders(headers)));

  Map<String, String> _makeHeaders(Map<String, String> headers) =>
      {}..addAll(defaultHeaders)..addAll(headers);

  Future<Document> _exec(Future<http.Response> fn(http.Client client)) async {
    final client = clientFactory();
    try {
      final response = await fn(client);
      return _makeDocument(response);
    } finally {
      client.close();
    }
  }

  Document _makeDocument(http.Response response) =>
      Document.fromJson(json.decode(response.body));
}

typedef http.Client ClientFactory();

abstract class AuthorizationHeader {
  final key = 'Authorization';

  String get value;

  Map<String, String> get asMap => {key: value};
}

class AuthorizationHeaderBasic extends AuthorizationHeader {
  AuthorizationHeaderBasic(String username, String password)
      : value = base64.encode(utf8.encode('Basic ${username}:${password}'));

  final String value;
}

class AuthorizationHeaderBearer extends AuthorizationHeader {
  AuthorizationHeaderBearer(String token) : value = 'Bearer ${token}';

  final String value;
}
